import os
import psutil
import logging
import pty
import fcntl
import struct
import termios
import asyncio
import subprocess
import re
import json
import select
from channels.generic.websocket import AsyncWebsocketConsumer

logger = logging.getLogger(__name__)

class StatsConsumer(AsyncWebsocketConsumer):
    # Статический список для отслеживания активных задач
    active_tasks = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.process = None
        self.master_fd = None
        self.stream_task = None
        self.keep_running = False
        self.session_start_time = None 

    async def connect(self):
        # Если уже есть активные подключения, закрываем их
        if StatsConsumer.active_tasks:
            for consumer in StatsConsumer.active_tasks:
                logger.info("Closing existing StatsConsumer connection.")
                await consumer.close(code=4000)
            StatsConsumer.active_tasks.clear()

        await self.accept()
        logger.info("WebSocket connection accepted.")
        self.keep_running = True
        self.session_start_time = None

        # Добавляем текущее подключение в список активных задач
        StatsConsumer.active_tasks.append(self)

        # Проверка, не запущен ли уже процесс KelVPNCli
        if self.is_process_running("KelVPNCli"):
            logger.error("KelVPNCli is already running. Terminating existing process.")
            self.terminate_existing_process("KelVPNCli")

        # Создаём PTY
        self.master_fd, slave_fd = pty.openpty()

        # Устанавливаем размер терминала
        desired_width = 200
        desired_height = 24
        fcntl.ioctl(
            slave_fd,
            termios.TIOCSWINSZ,
            struct.pack("HHHH", desired_height, desired_width, 0, 0),
        )
        
        # Запускаем процесс в PTY
        self.process = subprocess.Popen(
            ["sudo", "/opt/kelvpn/bin/KelVPNCli", "-T"],
            stdout=slave_fd,
            stderr=slave_fd,
            close_fds=True,
        )

        os.close(slave_fd)

        # Устанавливаем неблокирующий режим для master_fd
        flags = fcntl.fcntl(self.master_fd, fcntl.F_GETFL)
        fcntl.fcntl(self.master_fd, fcntl.F_SETFL, flags | os.O_NONBLOCK)

        # Запускаем задачу для стриминга статистики
        self.stream_task = asyncio.create_task(self.stream_stats())

    async def disconnect(self, close_code):
        logger.info("WebSocket connection closed.")
        self.keep_running = False

        # Убираем текущий экземпляр из списка активных задач
        if self in StatsConsumer.active_tasks:
            StatsConsumer.active_tasks.remove(self)

        # Отменяем фоновую задачу, если она существует
        if self.stream_task:
            self.stream_task.cancel()

        # Завершаем процесс, если он был запущен
        if self.process:
            self.process.terminate()
            self.process.wait()

        # Закрываем master_fd, если он открыт
        if self.master_fd is not None:
            try:
                os.close(self.master_fd)
                self.master_fd = None
            except OSError as e:
                logger.warning(f"Attempted to close master_fd, but it was already closed: {str(e)}")

    def is_process_running(self, process_name):
        """Проверяет, запущен ли процесс с указанным именем."""
        for proc in psutil.process_iter(attrs=["pid", "name"]):
            if process_name.lower() in proc.info["name"].lower():
                logger.debug(f"Found running process: {proc.info}")
                return True
        return False

    def terminate_existing_process(self, process_name):
        """Завершает процесс с указанным именем."""
        for proc in psutil.process_iter(attrs=["pid", "name"]):
            if process_name.lower() in proc.info["name"].lower():
                logger.info(f"Terminating process {proc.info['name']} (PID: {proc.info['pid']})")
                proc.terminate()
                try:
                    proc.wait(timeout=5)
                except psutil.TimeoutExpired:
                    logger.warning(f"Force killing process {proc.info['name']} (PID: {proc.info['pid']})")
                    proc.kill()

    async def stream_stats(self):
        try:
            while self.keep_running:
                await asyncio.sleep(1)
                ready, _, _ = select.select([self.master_fd], [], [], 0)
                if ready:
                    try:
                        data = os.read(self.master_fd, 4096)
                    except OSError:
                        continue

                    if not data:
                        break
                    decoded_data = data.decode("utf-8", errors="replace")
                    logger.debug(f"decoded_data: {decoded_data}")
                    lines = decoded_data.splitlines()
                    logger.debug(f"lines: {lines}")
                    for line in lines:
                        if line.strip():
                            if "uptime" not in line.lower():
                                continue
                            
                            logger.debug(f"line: {line}")
                            parsed_stats = self.parse_stats(line.strip())
                            try:
                                await self.send(text_data=json.dumps(parsed_stats))
                                logger.debug(f"Sent: {parsed_stats}")
                            except Exception as e:
                                logger.error(f"Error sending data: {str(e)}")
                                self.keep_running = False
                                break

        except asyncio.CancelledError:
            logger.info("Stream stats task was cancelled.")
        except Exception as e:
            logger.error(f"Error streaming stats: {str(e)}")
            try:
                await self.send(text_data=json.dumps({"error": "Error streaming stats."}))
            except Exception:
                pass

    def parse_stats(self, line):
        download_match = re.search(r"download\s+([\d.]+\s+\w+)", line, re.IGNORECASE)
        upload_match = re.search(r"upload\s+([\d.]+\s+\w+)", line, re.IGNORECASE)
        received_match = re.search(r"([\d.]+\s+\w+)\s+bytes received", line, re.IGNORECASE)
        sent_match = re.search(r"([\d.]+\s+\w+)\s+bytes sent", line, re.IGNORECASE)
        packets_received_match = re.search(r"(\d+)\s+packets received", line, re.IGNORECASE)
        packets_sent_match = re.search(r"(\d+)\s+packets sent", line, re.IGNORECASE)
        uptime_match = re.search(r"((?:\d+d-)?[\d:]+)\s+uptime", line, re.IGNORECASE)

        raw_uptime = uptime_match.group(1) if uptime_match else "0:00"

        # Пересчитываем время относительно первого полученного uptime
        corrected_uptime = self.recalculate_uptime(raw_uptime)

        return {
            "download": download_match.group(1) if download_match else "0 Bytes",
            "upload": upload_match.group(1) if upload_match else "0 Bytes",
            "received": received_match.group(1) if received_match else "0 Bytes",
            "sent": sent_match.group(1) if sent_match else "0 Bytes",
            "packets_received": int(packets_received_match.group(1)) if packets_received_match else 0,
            "packets_sent": int(packets_sent_match.group(1)) if packets_sent_match else 0,
            "uptime": corrected_uptime,
        }

    def recalculate_uptime(self, uptime_str: str) -> str:
        logger.debug(f"Raw uptime_str: {uptime_str}")
        if "d-" in uptime_str:
            _, time_part = uptime_str.split("d-")
        else:
            time_part = uptime_str

        logger.debug(f"Time part after split: {time_part}")
        try:
            time_parts = list(map(int, time_part.split(":")))
            logger.debug(f"Time parts: {time_parts}")
        except Exception as e:
            logger.error(f"Error converting uptime '{time_part}': {e}")
            return "0:00:00"

        if len(time_parts) == 3:
            hours, minutes, seconds = time_parts
        elif len(time_parts) == 2:
            hours = 0
            minutes, seconds = time_parts
        else:
            logger.error(f"Unexpected time format: {time_part}")
            return "0:00:00"

        total_seconds = hours * 3600 + minutes * 60 + seconds
        logger.debug(f"Total seconds: {total_seconds}")

        if self.session_start_time is None or total_seconds < self.session_start_time:
            self.session_start_time = total_seconds
            logger.debug(f"Reset session_start_time to {self.session_start_time}")
            return "00:00:00"

        elapsed_seconds = total_seconds - self.session_start_time
        corrected_hours = elapsed_seconds // 3600
        corrected_minutes = (elapsed_seconds % 3600) // 60
        corrected_seconds = elapsed_seconds % 60

        corrected_time = f"{corrected_hours:02}:{corrected_minutes:02}:{corrected_seconds:02}"
        logger.debug(f"Corrected time: {corrected_time}")
        return corrected_time
